#include <assert.h>
#include "NMathFtns.h"
#include "NHashedString.h"
#include "NText.h"
#include "NUtilDefs.h"

NHashedStringTable* NHashedStringTable::Instance = NULL;

NHashedString gNullStr("NULL");

NHashedStringTable* NHashedStringTable::GetInstance()
{
	return Instance ? Instance : (Instance = new NHashedStringTable());
}

void NHashedStringTable::DestroyInstance()
{
	SafeDelete(Instance);
}

NHashedString::NHashedString()
{
	id = gNullStr.getID();

#ifdef _DEBUG
	// For debugging ONLY, include a link back to the table
	DEBUG_string = getString();
	DEBUG_bCaseSensitive = false;
#endif
}

NHashedString::NHashedString(const char* string, bool bCaseSensitive)
{
	if (!string)
	{
		id = gNullStr.getID();
	}
	else
	{
		// We use a case-insensitive hash for all string regardless of desired case sensitivity.
		// This ensures that "foo" and "FOO" will be indexed identically.
		// When we put it in the table, we use the desired sensitivity to check whether we've found a match.
		// In this way, "FOO" can be specified to not match "foo" such that they will both end up here.
		// Where this gets weird is that I will then have two entries in the table, and each will have a different ID.
		// That means that "FOO" and "foo" will not meet the == operator check.
		// Maybe that's fine? I'm not sure yet.
		// Case insensitive strings will always hit the first match in the table, so they'll behave as before.
		// Only if we explicitly request case sensitivity will we get a different string back.

		// I can imagine this might get confusing in some cases.

		// First case:
		// I add "foo" first. Doesn't matter whether it's case sensitive or not. It's first.
		// I add "foo" insensitively. It matches and I get back the ID of the "foo" entry.
		// I add "foo" sensitively. It matches and I get back the ID of the "foo" entry.
		// I add "FOO" insensitively. It matches and I get back the ID of the "foo" entry.
		// I add "FOO" sensitively. It doesn't match, and I add the new entry.
		// Now the only way I can get a match on "FOO" is to request "FOO" sensitively.

		// Second case:
		// Let's say I add "FOO" first. Doesn't matter whether it's case sensitive or not. It's first.
		// I add "foo" insensitively. It matches and I get back the ID of the "FOO" entry.
		// I add "foo" sensitively. It doesn't match, and I add the new entry.
		// I add "FOO" insensitively. It matches and I get back the ID of the "FOO" entry.
		// I add "FOO" sensitively. It matches and I get back the ID of the "FOO" entry.
		// Now the only way I can get a match on "foo" is to request "foo" sensitively.

		// So whichever one went in first is the "real" one. That's not too bad, I suppose. That's how this normally works.
		// It'll only really be a problem if I'm trying to match against a string that was requested with case sensitivity
		// and I'm not getting a match and it's not clear why. An easy solution (albeit expensive) check would be
		// to change (A == B) to (NHashedString(A.getString()) == NHashedString(B.getString())) to force case insensitivity on both.
		// I suppose I could add the ^ operator for this case to keep parity with NString.
		// It might be worth flagging case-sensitive strings with a bool so I can tell what they are (at least when debugging).

		uint hash = GetCaseInsensitiveTextHashValue(string);
		hash %= HASHED_STRING_TABLE_SIZE;

		int instance = NHashedStringTable::GetInstance()->add(hash, string, bCaseSensitive);

		// 10 hash / 22 instance
		//id = (((hash << ID_INSTANCE_BITS) & ID_HIGH_10_MASK) + (instance & ID_LOW_22_MASK));

		// 22 instance / 10 hash
		id = (((instance << ID_HASH_BITS) & ID_HIGH_22_MASK) + (hash & ID_LOW_10_MASK));
	}

#ifdef _DEBUG
	// For debugging ONLY, include a link back to the table
	DEBUG_string = getString();
	DEBUG_bCaseSensitive = bCaseSensitive;
#endif
}

// TODO: Make this safe.  Return a copy?
const char* NHashedString::getString() const
{
	const char* retVal = NULL;

	// 10 hash / 22 instance
	//uint hash = ((id & ID_HIGH_10_MASK) >> ID_INSTANCE_BITS);
	//int instance = (id & ID_LOW_22_MASK);

	// 22 instance / 10 hash
	int instance = ((id & ID_HIGH_22_MASK) >> ID_HASH_BITS);
	uint hash = (id & ID_LOW_10_MASK);

	retVal = NHashedStringTable::GetInstance()->getString(hash, instance);

	return retVal;
}

// TODO: Rewrite, one entry, one exit.
int NHashedStringTable::add(int index, const char* string, bool bCaseSensitive)
{
	GetMutex(hHashTableMutex);

	// Maybe I might WANT an empty string, hurr?
	//assert(string && strlen(string) > 0 && index >= 0 && index < HASHED_STRING_TABLE_SIZE);
	assert(string && index >= 0 && index < HASHED_STRING_TABLE_SIZE);

	int retVal = 0;

	NHashedStringEntry* entry = entryArray[index];

	if (!entry)
	{
		NHashedStringEntry* newEntry = new NHashedStringEntry(string);
		entryArray[index] = newEntry;
	}
	else
	{
		bool bFound = false;

		while (!bFound)
		{
			if ((!bCaseSensitive && _stricmp(string, entry->getString()) == 0) ||
				(bCaseSensitive && strcmp(string, entry->getString()) == 0))
			{
				bFound = true;
			}
			else
			{
				retVal++;

				if (entry->next)
				{
					entry = entry->next;
				}
				else
				{
					entry->next = new NHashedStringEntry(string);
					bFound = true;
				}
			}
		}
	}

	ReleaseMutex(hHashTableMutex);
	return retVal;
}

// TODO: Rewrite, one entry, one exit.
const char* NHashedStringTable::getString(int index, int instance)
{
	GetMutex(hHashTableMutex);

	assert(index >= 0 && index < HASHED_STRING_TABLE_SIZE);

	const char* retVal = NULL;

	NHashedStringEntry* entry = entryArray[index];

	if (entry)
	{
		for (int i = 0; i < instance; ++i)
		{
			if (entry->next)
			{
				entry = entry->next;
			}
			else
			{
				return NULL;
			}
		}

		retVal = entry->getString();
	}

	ReleaseMutex(hHashTableMutex);

	return retVal;
}

NHashedStringTable::NHashedStringTable()
{
	hHashTableMutex = (HANDLE)CreateMutex(NULL, FALSE, NULL);

	memset(entryArray, 0, HASHED_STRING_TABLE_SIZE * sizeof(NHashedStringEntry*));
}

NHashedStringTable::~NHashedStringTable()
{
	for (int i = 0; i < HASHED_STRING_TABLE_SIZE; ++i)
	{
		if (entryArray[i])
		{
			recCleanup(entryArray[i]);
		}
	}

	CloseHandle(hHashTableMutex);
}

void NHashedStringTable::recCleanup(NHashedStringEntry* entry)
{
	assert(entry);

	if (entry->next)
	{
		recCleanup(entry->next);
	}
	SafeDelete(entry);
}

void NHashedStringTable::ReportStats(int& ArraySize, int& StringCount, int& RowMin, int& RowMax, float& Mean, float& Variance, float& StdDev)
{
	ArraySize = HASHED_STRING_TABLE_SIZE;
	StringCount = 0;
	Mean = 0.0f;
	Variance = 0.0f;
	StdDev = 0.0f;
	RowMin = 1000000;
	RowMax = 0;

	for (int i = 0; i < HASHED_STRING_TABLE_SIZE; ++i)
	{
		NHashedStringEntry* entry = entryArray[i];
		while (entry)
		{
			StringCount++;
			entry = entry->next;
		}
	}

	Mean = float(StringCount) / float(ArraySize);

	for (int i = 0; i < HASHED_STRING_TABLE_SIZE; ++i)
	{
		int ThisCount = 0;
		NHashedStringEntry* entry = entryArray[i];
		while (entry)
		{
			ThisCount++;
			entry = entry->next;
		}

		RowMax = Max(RowMax, ThisCount);
		RowMin = Min(RowMin, ThisCount);

		Variance += ((float(ThisCount) - Mean) * (float(ThisCount) - Mean));
	}
	Variance /= float(ArraySize);

	StdDev = sqrt(Variance);
}

NHashedStringEntry::NHashedStringEntry(const char* _string)
{
	// Maybe I might WANT an empty string, hurr?
	//assert(_string && strlen(_string) > 0);
	assert(_string);

	int len = Min(MAX_HASHED_STRING_LENGTH-1, int(strlen(_string)));

	string = new char[len+1];

	memcpy_s(string, len, _string, len);

	string[len] = 0;

	next = NULL;
}

NHashedStringEntry::~NHashedStringEntry()
{
	SafeDeleteArray(string);
}
