#pragma once

#include <stdlib.h>
#include <math.h>
#include "NMathDefs.h"

static inline int Max(int i, int j) {return (i > j) ? i : j;}
static inline int Min(int i, int j) {return (i < j) ? i : j;}
static inline int Clamp(int num, int min, int max) {return Min(Max(num, min), max);}
static inline float Max(float i, float j) {return (i > j) ? i : j;}
static inline float Min(float i, float j) {return (i < j) ? i : j;}
static inline float Clamp(float num, float min, float max) {return Min(Max(num, min), max);}
static inline float Saturate(float num) {return Min(Max(num, 0.0f), 1.0f);}

// Knuth's "essentially-equal" function.  "Approximately-equal" takes the max of the absolutes.
static inline bool IsEqual(float s, float t, float tolerance = 0.00001f)
{
	return ((float)fabs(s - t) <= tolerance * Min((float)fabs(s), (float)fabs(t)));
}

// NERD: This is almost never an ideal solution.  Assuming I'm wanting to do this to more than one i,
// I'd want to store off the reciprocal of (max - min) and multiply.
static inline float Normalize(float i, float min, float max) {return (i - min) / (max - min);}

template <class T> void Swap(T& i, T& j)
{
	T temp = i;
	i = j;
	j = temp;
}

// NERD: Look into alternative (faster or better) random number generators.
// NERD: Having an API like this is good, though.  I can change the implementation discreetly.
static const float InvRandMax = 1.0f / float(RAND_MAX);
static const float InvRandMaxPlusOne = 1.0f / (float(RAND_MAX) + 1);

static inline int random_fl(int range, int minimum = 0) {return int(((float(rand()) + 0.5f) * InvRandMaxPlusOne) * float(range)) + minimum;}
static inline float random_fl() {return float(rand()) * InvRandMax;}					// one inclusive
static inline float random_1inc() {return random_fl();}
static inline float random_1exc() {return float(rand()) * InvRandMaxPlusOne;}		// one exclusive
static inline float random_0exc() {return float(rand() + 1) * InvRandMaxPlusOne;}		// zero exclusive
static inline float random_01exc() {return (float(rand()) + 0.5f) * InvRandMaxPlusOne;}		// zero and one exclusive

static inline float Lerp(float x1, float x2, float dist) {return x1 + dist * (x2 - x1);}
static inline float Bilerp(float x1y1, float x2y1, float x1y2, float x2y2, float xDist, float yDist)
		{return ((1.0f - yDist) * Lerp(x1y1,x2y1, xDist)) + (yDist * Lerp(x1y2,x2y2, xDist));}

static inline float CosLerp(float x1, float x2, float dist) {return (((cos(dist*PI)+ 1.0f) * 0.5f) * x1) + ((1.0f - ((cos(dist*PI)+ 1.0f) * 0.5f)) * x2);}
static inline float CosBilerp(float x1y1, float x2y1, float x1y2, float x2y2, float xDist, float yDist)
		{return (((cos(yDist*PI)+ 1.0f) * 0.5f) * CosLerp(x1y1,x2y1, xDist)) + ((1.0f - ((cos(yDist*PI)+ 1.0f) * 0.5f)) * CosLerp(x1y2,x2y2, xDist));}

static inline float Cubic(float t)
{
	return (t*t)*(3.0f - (2.0f * t));
}

static inline float Quintic(float t)
{
	return (t*t*t)*(t*(t*6.0f - 15.0f) + 10.0f);
}

// Not actually a cubic!  This is a quintic.  It has a continuous second derivative.
// (Taken from one of Ken Perlin's improvements to Perlin noise.)
static inline float CubicLerp(float a, float b, float t)
{
	return a + (t*t*t)*(t*(t*6.0f - 15.0f) + 10.0f)*(b-a);

	// Actual cubic:
	//return a + (t*t)*(3.0f - (2.0f * t))*(b-a);
}

static inline float CubicBilerp(float x0y0, float x1y0, float x0y1, float x1y1, float tx, float ty)
{
	return CubicLerp(CubicLerp(x0y0, x1y0, tx), CubicLerp(x0y1, x1y1, tx), ty);
}

// NERD: Have a FULL cubic interpolation (four points in one dimension, sixteen in two).
// NERD: It's a pretty simple formula, and it's all over the webatrons.  The simplified form (-2t^3 + 3t) assumes zero tangents.

static inline nlong LongEndianSwap(nlong num)
{
	return ((num & 0x000000ff) << 24) |
		   ((num & 0x0000ff00) << 8) |
		   ((num & 0x00ff0000) >> 8) |
		   ((num & 0xff000000) >> 24);
}

static inline ushort ShortEndianSwap(ushort num)
{
	return ((num & 0x00ff) << 8) |
		   ((num & 0xff00) >> 8);
}

static inline bool IsLittleEndian()
{
	int test = 1;
	return (*((char*)&test) == 1);
}

static inline bool IsBigEndian() {return !IsLittleEndian();}

// I can add more like this later if I need them...this makes endian swapping on a float quicker and easier
static inline float EndianSwap(float num)
{
	// TODO: Assert float and nlong are same length!!
	nlong swapped = LongEndianSwap(*(nlong*)&num);
	return *(float*)&swapped;
}

// NERD: I think this is pretty recent, and it looks pretty fast.  Keep it.
// NERD: I guess I could unroll the loop, but HONESTLY.  =/
// 02/08/2010: Another way to do this: Log2(x) = Log10(x) / Log10(2.0f)
static inline int LogBaseTwo(int i)
{
	int shifts = -1;

	while (i > 0)
	{
		i >>= 1;
		++shifts;
	}

	return shifts;
}

static inline float LogBase(float x, float base)
{
	return log(x) / log(base);
}

static inline float log2(float x)
{
	return log(x) / log(2.0f);
}

// Probability of getting x, given mean and variance
static inline float NormalProbability(float x, float mean, float variance)
{
	return (1.0f / (sqrt(variance) * sqrt(2.0f * PI))) * exp(-((x-mean)*(x-mean)) / (2.0f * variance));
}

// Returns a random value from a normal distribution given a mean and STANDARD DEVIATION
// (99.7300% of values are within three standard deviations of the mean,
// so given a desired range R, std. dev. ~=  R / 6)
// In practice, this also never seems to return values beyond 2R (or R/6 on either side of the mean),
// so for safety purposes, a range of 2R may be assumed. (Tested over a billion samples. May run further when using an RNG with a larger max.)
static inline float RandomNormal(float mean, float stddev)
{
	float x = 0.0f;
	float y = 0.0f;
	float s = 2.0f;

	while (s >= 1.0f || s == 0.0f)
	{
		x = random_fl() * 2.0f - 1.0f;
		y = random_fl() * 2.0f - 1.0f;
		s = x*x + y*y;
	}

	// Implicit in "x * mul" or "y * mul" is the division by sqrt(s),
	// which projects the point (x,y) onto the circumference of the unit circle.

	// So if we say CHI = sqrt(-2 ln s) then the return value is
	// CHI * x/sqrt(s) * stddev + mean
	// And x/sqrt(s) is ALSO sin(TAU * U), where U is a uniform random.
	// So this is
	// sqrt(-2 ln U1) * sin(TAU * U) * stddev + mean

	// mul is actually the product of this scalar and the chi variate sqrt(-2 ln(s)), where s is a uniformly distributed number (0,1].

	// The part of this function that's more difficult to cut out is generating the (x,y) pair, or reducing this to a single number.
	// I want a function that returns me a value [-1,+1] with the appropriate weighting. This weighting is the tricky part.
	// Maybe dart throwing is the best option?

	float mul = sqrt(-2.0f * log(s) / s);
	return x * mul * stddev + mean;		// y * mul * stddev + mean is another perfectly valid return value.  it gets thrown away.  shrug.
}

static inline int sign(int x)
{
	int RetVal = 0;

	if (x > 0)
	{
		RetVal = 1;
	}
	else if (x < 0)
	{
		RetVal = -1;
	}

	return RetVal;
}

static inline float sign(float x)
{
	float RetVal = 0.0f;

	if (x > 0.0f)
	{
		RetVal = 1.0f;
	}
	else if (x < 0.0f)
	{
		RetVal = -1.0f;
	}

	return RetVal;
}

void DCT(float* complexVals, int numVals);
void DCTi(float* complexVals, int numVals);
void DCT2(float** complexVals, int numValsX, int numValsY);
void DCT2i(float** complexVals, int numValsX, int numValsY);

// Essentially models a simple spring with Euler integration.
// The distance we move is proportional to the distance to our target.
static inline float Smooth(float current, float target, float speed, float deltaTime)
{
	return Lerp(current, target, Clamp(deltaTime * speed, 0.0f, 1.0f));
}

static const int SINE_TABLE_SIZE = 256;

static const float sintable[SINE_TABLE_SIZE] =
{
	0.000000f,0.024541f,0.049068f,0.073565f,0.098017f,0.122411f,0.146730f,0.170962f,0.195090f,0.219101f,0.242980f,0.266713f,0.290285f,0.313682f,0.336890f,0.359895f,
	0.382683f,0.405241f,0.427555f,0.449611f,0.471397f,0.492898f,0.514103f,0.534998f,0.555570f,0.575808f,0.595699f,0.615232f,0.634393f,0.653173f,0.671559f,0.689541f,
	0.707107f,0.724247f,0.740951f,0.757209f,0.773010f,0.788346f,0.803208f,0.817585f,0.831470f,0.844854f,0.857729f,0.870087f,0.881921f,0.893224f,0.903989f,0.914210f,
	0.923880f,0.932993f,0.941544f,0.949528f,0.956940f,0.963776f,0.970031f,0.975702f,0.980785f,0.985278f,0.989177f,0.992480f,0.995185f,0.997290f,0.998795f,0.999699f,
	1.000000f,0.999699f,0.998795f,0.997290f,0.995185f,0.992480f,0.989177f,0.985278f,0.980785f,0.975702f,0.970031f,0.963776f,0.956940f,0.949528f,0.941544f,0.932993f,
	0.923880f,0.914210f,0.903989f,0.893224f,0.881921f,0.870087f,0.857729f,0.844854f,0.831470f,0.817585f,0.803208f,0.788346f,0.773010f,0.757209f,0.740951f,0.724247f,
	0.707107f,0.689541f,0.671559f,0.653173f,0.634393f,0.615232f,0.595699f,0.575808f,0.555570f,0.534997f,0.514103f,0.492898f,0.471397f,0.449611f,0.427555f,0.405241f,
	0.382683f,0.359895f,0.336890f,0.313682f,0.290285f,0.266713f,0.242980f,0.219101f,0.195090f,0.170962f,0.146730f,0.122411f,0.098017f,0.073564f,0.049067f,0.024541f,
	-0.000000f,-0.024541f,-0.049068f,-0.073565f,-0.098017f,-0.122411f,-0.146730f,-0.170962f,-0.195090f,-0.219101f,-0.242980f,-0.266713f,-0.290285f,-0.313682f,-0.336890f,-0.359895f,
	-0.382683f,-0.405241f,-0.427555f,-0.449612f,-0.471397f,-0.492898f,-0.514103f,-0.534998f,-0.555570f,-0.575808f,-0.595699f,-0.615232f,-0.634393f,-0.653173f,-0.671559f,-0.689541f,
	-0.707107f,-0.724247f,-0.740951f,-0.757209f,-0.773010f,-0.788346f,-0.803208f,-0.817585f,-0.831470f,-0.844854f,-0.857729f,-0.870087f,-0.881921f,-0.893224f,-0.903989f,-0.914210f,
	-0.923880f,-0.932993f,-0.941544f,-0.949528f,-0.956940f,-0.963776f,-0.970031f,-0.975702f,-0.980785f,-0.985278f,-0.989177f,-0.992480f,-0.995185f,-0.997290f,-0.998795f,-0.999699f,
	-1.000000f,-0.999699f,-0.998795f,-0.997290f,-0.995185f,-0.992480f,-0.989177f,-0.985278f,-0.980785f,-0.975702f,-0.970031f,-0.963776f,-0.956940f,-0.949528f,-0.941544f,-0.932993f,
	-0.923879f,-0.914210f,-0.903989f,-0.893224f,-0.881921f,-0.870087f,-0.857729f,-0.844853f,-0.831469f,-0.817585f,-0.803208f,-0.788346f,-0.773010f,-0.757209f,-0.740951f,-0.724247f,
	-0.707107f,-0.689541f,-0.671559f,-0.653173f,-0.634393f,-0.615231f,-0.595699f,-0.575808f,-0.555570f,-0.534998f,-0.514103f,-0.492898f,-0.471397f,-0.449611f,-0.427555f,-0.405241f,
	-0.382683f,-0.359895f,-0.336890f,-0.313682f,-0.290284f,-0.266712f,-0.242980f,-0.219101f,-0.195090f,-0.170962f,-0.146730f,-0.122410f,-0.098017f,-0.073565f,-0.049068f,-0.024541f,
};

static const float RadsToSineTableIndex = float(SINE_TABLE_SIZE) / (2.0f * PI);

// Not safe for negative numbers.
static inline float fastsin(float x)
{
	x *= RadsToSineTableIndex;
	int intpart = (int)x;
	float decpart = x - (float)intpart;

	int x1 = (intpart)%SINE_TABLE_SIZE;
	int x2 = (intpart+1)%SINE_TABLE_SIZE;

	float s1 = sintable[x1];
	float s2 = sintable[x2];

	return (s1*(1.0f-decpart))+(s2*decpart);
}

// Not safe for negative numbers.
static inline float fastcos(float x)
{
	return fastsin(x+HALF_PI);
}

static inline uint GreatestCommonDivisor(uint a, uint b)
{
	if (b == 0)
	{
		return a;
	}
	else
	{
		return GreatestCommonDivisor(b, a % b);
	}
}

// Quake's fast inverse square root
static inline float FastInvSqrt(float x)
{
	float xhalf = 0.5f * x;
	int i = *(int*)&x;
	i = 0x5f3759d5 - (i >> 1);		// alternate value: 0x5f375a86 (supposedly more accurate)
	x = *(float*)&i;
	x = x*(1.5f - xhalf*x*x);
	return x;
}

float HorizontalFOV(float VerticalFOV, float AspectRatio);
float VerticalFOV(float HorizontalFOV, float AspectRatio);

unsigned int atox(const char* InText);
