#pragma once

#include <memory.h>
#include <stdlib.h>
#include <string.h>
#include "NMathDefs.h"
#include "NHashedString.h"
#include <vector>

// 03/03/2011: Adding length variable is a bummer because previously sizeof(NString) was 4 (one char*). Oh well...

// 02/04/2011: Adding the option for this to contain arbitrary inlined image data.
// This will break strlen()/Length(), etc.

// This needs to be both bulletproof and full-featured.
// There should be absolutely no access to the actual char*.
// Copy constructor and assignment operator must do deep copies.
// Addition operator should concatenate.
// Additional features such as getting left, right, and middle sections may be useful, as well.

// Restricting access to the char* is going to present some problems, such as passing a char*
// to an effect to set a texture name.  The best solution I can think of is to have a function
// which takes in a pointer reference and copies the string to it.

// No, wait, that's stupid.  That would make it really easy to accidentally leak memory.
// I think it's going to have to be able to return the actual pointer.

struct NInlinedImage
{
	NHashedString SetName;
	NHashedString GlyphName;
};

enum EUTF8Presence
{
	EU8_Unknown,
	EU8_No,
	EU8_Yes,
};

class NString
{
public:
	NString();
	NString(const NString& string);
	NString(const NHashedString& string);
	NString(char* string, int ExactLength=-1);
	NString(const char* string, int ExactLength=-1);
	~NString()
	{
		free();
	}

	char* operator *()	const {return text;}
	int Length()		const {return length;}
	int PercLen()		const {return prcvln;}
	int PercToRealLen(int InPercLen) const;					// Takes in a desired perceived substring length and returns the number of bytes required to draw that much text, accounting for glyphs and UTF-8 characters.

	NString operator =(const NString& string);
	NString operator =(const NHashedString& string);
	NString operator =(char* string);
	NString operator =(const char* string);

	NString operator +=(const NString& string);
	NString operator +=(const NHashedString& string);
	NString operator +=(char* string);
	NString operator +=(const char* string);

	NString operator +(const NString& string) const;
	NString operator +(const NHashedString& string) const;
	NString operator +(char* string) const;
	NString operator +(const char* string) const;

	bool operator ==(const NString& string) const;
	bool operator ==(const NHashedString& string) const;
	bool operator ==(char* string) const;
	bool operator ==(const char* string) const;

	// Ugly and arbitrary, but we'll say this is an "approximately equal to" operator since I can't add a "~=" like I really want.
	bool operator ^(const NString& string) const;
	bool operator ^(const NHashedString& string) const;
	bool operator ^(char* string) const;
	bool operator ^(const char* string) const;

	bool operator !=(const NString& string) const;
	bool operator !=(const NHashedString& string) const;
	bool operator !=(char* string) const;
	bool operator !=(const char* string) const;

	int Copy(char*& OutText) const;

	NString Left(int InCount) const;
	NString Right(int InCount) const;
	NString Mid(size_t Pos, size_t Count) const;		// indexed from 0, obviously
	int Split(const NString& Partition, std::vector<NString>& OutArray) const;
	NString Replace(const NString& SearchTerm, const NString& ReplacementTerm) const;		// Slow; uses Split() to separate and then rebuilds.
	int Find(const NString& Term, size_t From = 0) const;

	// Trim whitespace from both sides of the string.
	NString Trimmed() const;

	//int Find(const NString& SearchString);

	void ResolveAll();
	static NString ResolveEscapeChars(const NString& InString);
	static NString ResolveMarkup(const NString& InString);

	NString GetStripped(NString TestChars);

	NString GetStrippedNonalphanumeric();

	// Helper functions for source code parsing.
	NString StripBlockComments();
	NString StripLineComments();

	NString GetUpper() const;
	NString GetLower() const;

	int GetInt() const;
	float GetFloat() const;
	bool GetBool() const;

	inline bool operator <(const NString& other) const {return (strcmp(text, other.text) < 0);}
	inline bool operator <=(const NString& other) const {return (strcmp(text, other.text) <= 0);}
	inline bool operator >(const NString& other) const {return (strcmp(text, other.text) > 0);}
	inline bool operator >=(const NString& other) const {return (strcmp(text, other.text) >= 0);}

	EUTF8Presence GetUTF8Presence() {return utf8;}
	bool IsUTF8() {return (utf8 == EU8_Yes);}			// Eh. Really all strings are UTF-8 strings now, so this is more about whether it has those characters.

private:
	static char* text_null;		// Default empty string. Other empty strings (such as those created by assigning "" to an NString) are not guaranteed to point to this.
	char* text;
	int length;
	int prcvln;					// Perceived length, accounting for glyphs (treated as one character) and UTF-8 characters.
	EUTF8Presence utf8;

	void free();
	void init_empty();

	static void SafeStringDelete(char* InText);

	int CalcPercLen(int ExactLength=-1) const;
};
