#include <assert.h>
#include "NText.h"
#include "NUtilDefs.h"
#include "NString.h"
#include "NMathFtns.h"

NStringMgr* NStringMgr::Instance = NULL;
int NStringMgr::InitCount = 0;

NStringMgr::NStringMgr()
{
	InitCount++;
	assert(InitCount == 1);

	hTextMutex = (HANDLE)CreateMutex(NULL, FALSE, NULL);
}

NStringMgr::~NStringMgr()
{
	update();		// Do one final update before destroying
	CloseHandle(hTextMutex);
}

void NStringMgr::update()
{
	GetMutex(hTextMutex);

	for (int i = 0; i < (int)strings.size(); ++i)
	{
		delete[] strings[i];
	}

	while (!strings.empty())
	{
		strings.erase(strings.begin());
	}

	ReleaseMutex(hTextMutex);
}

inline void NStringMgr::push(char* string)
{
	GetMutex(hTextMutex);
	strings.push_back(string);
	ReleaseMutex(hTextMutex);
}

char* Text(const char* string, ...)
{
	int length;
	va_list args1;
	va_start(args1, string);

	length = _vscprintf(string, args1) + 1;
	char* newStr = new char[length];
	vsprintf_s(newStr, length, string, args1);

	// I wasn't doing this on Windows...
	va_end(args1);

	NStringMgr::GetInstance()->push(newStr);
	return newStr;
}

uint GetTextHashValue(const char* str)
{
	assert(str);

	uint hash = 5381;
	int c = *str;

	while (c)
	{
		hash = ((hash << 5) + hash) + c;
		str++;
		c = *str;
	}

	return hash;
}

uint GetCaseInsensitiveTextHashValue(const char* str)
{
	assert(str);

	uint hash = 5381;
	int c = *str;

	while (c)
	{
		if (c >= 65 && c <= 90)
		{
			c += 32;
		}

		hash = ((hash << 5) + hash) + c;
		str++;
		c = *str;
	}

	return hash;
}

uint GetTextHashValue_ELF(const char* str)
{
	assert(str);

	const char* cit = str;
	uint hash = 0;
	uint x;
	char c = *cit;
	while (c)
	{
		hash = (hash << 4) + c;
		if ((x = hash & 0xf0000000) != 0)
		{
			hash ^= (x >> 24);
		}
		hash &= ~x;

		cit++;
		c = *cit;
	}

	return hash;
}

char* StringCopy(char* InString, int ExactLength)
{
	char* RetVal = NULL;

	if (InString)
	{
		size_t Len = (ExactLength < 0) ? strlen(InString) : (size_t)Min(ExactLength, (int)strlen(InString));
		RetVal = new char[Len+1];
		memcpy_s(RetVal, Len * sizeof(char), InString, Len * sizeof(char));
		RetVal[Len] = 0;
	}

	return RetVal;
}

char* StringCopy(const char* InString, int ExactLength)
{
	char* RetVal = NULL;

	if (InString)
	{
		size_t Len = (ExactLength < 0) ? strlen(InString) : (size_t)Min(ExactLength, (int)strlen(InString));
		RetVal = new char[Len+1];
		memcpy_s(RetVal, Len * sizeof(char), InString, Len * sizeof(char));
		RetVal[Len] = 0;
	}

	return RetVal;
}

char* StringCat(char* InString1, char* InString2)
{
	char* RetVal = NULL;

	if (InString1 && InString2)
	{
		size_t Len1 = strlen(InString1);
		size_t Len2 = strlen(InString2);

		{
			RetVal = new char[Len1 + Len2 +1];

			if (Len1 > 0)
			{
				memcpy_s(RetVal, Len1 * sizeof(char), InString1, Len1 * sizeof(char));
			}

			if (Len2 > 0)
			{
				// Not sure this " + Len1" is correct
				memcpy_s(RetVal + Len1, Len2 * sizeof(char), InString2, Len2 * sizeof(char));
			}

			RetVal[Len1 + Len2] = 0;
		}
	}

	return RetVal;
}

char* StringCat(char* InString1, const char* InString2)
{
	char* RetVal = NULL;

	if (InString1 && InString2)
	{
		size_t Len1 = strlen(InString1);
		size_t Len2 = strlen(InString2);

		{
			RetVal = new char[Len1 + Len2 +1];

			if (Len1 > 0)
			{
				memcpy_s(RetVal, Len1 * sizeof(char), InString1, Len1 * sizeof(char));
			}

			if (Len2 > 0)
			{
				// Not sure this " + Len1" is correct
				memcpy_s(RetVal + Len1, Len2 * sizeof(char), InString2, Len2 * sizeof(char));
			}

			RetVal[Len1 + Len2] = 0;
		}
	}

	return RetVal;
}

char* StringCat(char* InString1, int StrLen1, char* InString2, int StrLen2)
{
	assert(StrLen1 >= 0 && StrLen2 >= 0);

	char* RetVal = NULL;

	if (InString1 && InString2)
	{
		{
			RetVal = new char[StrLen1 + StrLen2 +1];

			if (StrLen1 > 0)
			{
				memcpy_s(RetVal, StrLen1 * sizeof(char), InString1, StrLen1 * sizeof(char));
			}

			if (StrLen2 > 0)
			{
				// Not sure this " + Len1" is correct
				memcpy_s(RetVal + StrLen1, StrLen2 * sizeof(char), InString2, StrLen2 * sizeof(char));
			}

			RetVal[StrLen1 + StrLen2] = 0;
		}
	}

	return RetVal;
}

char* StringCat(char* InString1, int StrLen1, const char* InString2, int StrLen2)
{
	assert(StrLen1 >= 0 && StrLen2 >= 0);

	char* RetVal = NULL;

	if (InString1 && InString2)
	{
		{
			RetVal = new char[StrLen1 + StrLen2 +1];

			if (StrLen1 > 0)
			{
				memcpy_s(RetVal, StrLen1 * sizeof(char), InString1, StrLen1 * sizeof(char));
			}

			if (StrLen2 > 0)
			{
				// Not sure this " + Len1" is correct
				memcpy_s(RetVal + StrLen1, StrLen2 * sizeof(char), InString2, StrLen2 * sizeof(char));
			}

			RetVal[StrLen1 + StrLen2] = 0;
		}
	}

	return RetVal;
}

char* StripNonAlphanumeric_Alloc(char* InString)
{
	char* RetVal = NULL;

	size_t len = strlen(InString) + 1;
	if (len > 0)
	{
		char* OutString = new char[len];
		memset(OutString, 0, len);

		char* TestString = InString;
		int pos = 0;

		while (*TestString)
		{
			char testChar = *TestString;

			if (IsAlphanumeric(testChar))
			{
				OutString[pos] = ToLower(testChar);
				pos++;
			}

			TestString++;
		}

		OutString[pos] = 0;

		RetVal = OutString;
	}

	return RetVal;
}

char* StripNonAlphanumeric_Inline(char* InString)
{
	// Not really "in-line," since it just allocates it, overwrites the input string, and deletes the copy...but whatevs.
	// I can rewrite it to legitimately adjust the incoming string in line if it's too slow, but I don't expect it to be an issue.

	char* RetVal = NULL;

	size_t len = strlen(InString) + 1;
	if (len > 0)
	{
		char* OutString = StripNonAlphanumeric_Alloc(InString);

		memset(InString, 0, len);
		memcpy_s(InString, len, OutString, len);

		delete[] OutString;
	}

	return RetVal;
}

bool StringToBool(const char* string)
{
	opt_assert(string);

	bool boolValue = false;

	if ((_stricmp(string, "true") == 0) ||
		(_stricmp(string, "t") == 0) ||
		(_stricmp(string, "yes") == 0) ||
		(_stricmp(string, "y") == 0) ||
		(_stricmp(string, "on") == 0))
	{
		boolValue = true;
	}
	else if ((_stricmp(string, "false") == 0) ||
		(_stricmp(string, "f") == 0) ||
		(_stricmp(string, "no") == 0) ||
		(_stricmp(string, "n") == 0) ||
		(_stricmp(string, "off") == 0))
	{
		boolValue = false;
	}
	else
	{
		int intValue = atoi(string);
		boolValue = (intValue != 0);
	}

	return boolValue;
}

int RomanNumeralChars(int SingleDigit)
{
	opt_assert(SingleDigit >= 0 && SingleDigit <= 9);

	return CharsPerDigit[SingleDigit];
}

int TotalRomanNumeralChars(int i)
{
	int RetVal = 0;

	while (i > 0)
	{
		RetVal += RomanNumeralChars(i % 10);
		i /= 10;
	}

	return RetVal;
}

// Handles values between 1 and 3999 inclusive.
NString RomanNumeral(int i)
{
	static char Buffer[16] = {0};

	if (i > 0 && i < 4000)
	{
		int DigitsNeeded = TotalRomanNumeralChars(i);
		opt_assert(DigitsNeeded < 16);

		Buffer[DigitsNeeded] = 0;

		int DecimalPlace = 0;
		int StrPos = DigitsNeeded;
		while (i > 0)
		{
			int Digit = (i % 10);

			int Chars = RomanNumeralChars(Digit);

			StrPos -= Chars;

			for (int c = 0; c < Chars; ++c)
			{
				Buffer[StrPos + c] = RomanDigits[DecimalPlace][Digit][c];
			}

			i /= 10;
			DecimalPlace++;
		}
	}
	else
	{
		Buffer[0] = 0;
	}

	return NString(Buffer);
}

// Hex stuff
int HexCharToInt(char InChar)
{
	int RetVal = -1;

	if (InChar >= 48 && InChar <= 57)
	{
		RetVal = int(InChar) - 48;
	}
	else if (InChar >= 65 && InChar <= 70)
	{
		RetVal = int(InChar) - 65 + 10;
	}
	else if (InChar >= 97 && InChar <= 102)
	{
		RetVal = int(InChar) - 97 + 10;
	}

	opt_assert(RetVal >= 0);

	return RetVal;
}
