#pragma once

#include "NWindowsVersion.h"
#include "NUtilDefs.h"

#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <vector>

// string manager class
class NStringMgr
{
public:
	void update();
	~NStringMgr();

private:
	std::vector<char*> strings;
	inline void push(char* string);
	friend char* Text(const char* string, ...);

	HANDLE hTextMutex;

	// 10/06/2009: Singleton pattern implementation
private:
	static NStringMgr* Instance;

	// 12/27/2010: Counter used to ensure we don't reinitialize this by accident after shutting it down.
	// Might be beneficial for all singletons, but definitely for this one.
	static int InitCount;

	NStringMgr();
	NStringMgr(const NStringMgr&) {}

public:
	static NStringMgr* GetInstance()
	{
		return Instance ? Instance : (Instance = new NStringMgr());
	}

	static void DestroyInstance()
	{
		SafeDelete(Instance);
	}
};

// I do want this to be globally scoped.  It would be a pain if it weren't.
char* Text(const char* string, ...);

uint GetTextHashValue(const char* str);
uint GetCaseInsensitiveTextHashValue(const char* str);
uint GetTextHashValue_ELF(const char* str);					// For Buzzard hashing consistency.

// Lots of C&P code in these implementations
char* StringCopy(char* InString, int ExactLength=-1);
char* StringCopy(const char* InString, int ExactLength=-1);

// Lots of C&P code in these implementations
char* StringCat(char* InString1, char* InString2);
char* StringCat(char* InString1, const char* InString2);

// Lots of C&P code in these implementations
char* StringCat(char* InString1, int StrLen1, char* InString2, int StrLen2);
char* StringCat(char* InString1, int StrLen1, const char* InString2, int StrLen2);

// General purpose stringy-type functions, partially taken from the NERDScript lexer:

static inline bool IsANSI(char InChar)
{
	// This should also hold true for the first byte of UTF-8 sequences.

	return ((InChar >= 32 || InChar < 0) &&
		InChar != 127 &&		// delete character / inlined image flag
		InChar != -127 &&		// 129 (unspecified character)
		InChar != -115 &&		// 141 (unspecified character)
		InChar != -113 &&		// 143 (unspecified character)
		InChar != -112 &&		// 144 (unspecified character)
		InChar != -99);			// 157 (unspecified character)
}

static inline bool IsWhitespace(char InChar)
{
	return (InChar == ' ' ||
		InChar == '\n' ||
		InChar == '\r' ||
		InChar == '\t' ||
		InChar == '\v' ||
		InChar == '\f');
}

static inline bool IsAlphaUpper(char InChar)
{
	return (InChar >= 65 && InChar <= 90);
}

static inline bool IsAlphaLower(char InChar)
{
	return (InChar >= 97 && InChar <= 122);
}

static inline bool IsUnderscore(char InChar)
{
	return (InChar == 95);
}

static inline bool IsAlpha(char InChar)
{
	return (IsAlphaUpper(InChar) || IsAlphaLower(InChar) || IsUnderscore(InChar));
}

static inline bool IsNumeric(char InChar)
{
	return (InChar >= 48 && InChar <= 57);
}

static inline bool IsAlphanumeric(char InChar)
{
	return (IsAlpha(InChar) || IsNumeric(InChar));
}

static inline char ToUpper(char InChar)
{
	return (IsAlphaLower(InChar)) ? InChar - 32 : InChar;
}

static inline char ToLower(char InChar)
{
	return (IsAlphaUpper(InChar)) ? InChar + 32 : InChar;
}

char* StripNonAlphanumeric_Alloc(char* InString);
char* StripNonAlphanumeric_Inline(char* InString);

bool StringToBool(const char* string);

// Roman numeral stuff
class NString;
static int CharsPerDigit[10] = {0,1,2,3,2,1,2,3,4,2};

static char* RomanOnes[10] = {"", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX"};
static char* RomanTens[10] = {"", "X", "XX", "XXX", "XL", "L", "LX", "LXX", "LXXX", "XC"};
static char* RomanHundreds[10] = {"", "C", "CC", "CCC", "CD", "D", "DC", "DCC", "DCCC", "CM"};
static char* RomanThousands[10] = {"", "M", "MM", "MMM", "M?", "?", "?M", "?MM", "?MMM", "M?"};

static char** RomanDigits[4] = {RomanOnes, RomanTens, RomanHundreds, RomanThousands};

int RomanNumeralChars(int SingleDigit);
int TotalRomanNumeralChars(int i);
NString RomanNumeral(int i);

// Hex stuff
int HexCharToInt(char InChar);
