#pragma once

#include <math.h>
#include "NMathDefs.h"
#include "NMathFtns.h"

class NVec2;
class NVec4;
class NMat4;

class NVec3;
inline NVec3 Negate(const NVec3& v);

class NVec3
{
public:
	// Constructors
	NVec3() : x(0.0f), y(0.0f), z(0.0f) {}
	NVec3(float _x) : x(_x), y(_x), z(_x) {}
	NVec3(float _x, float _y, float _z) : x(_x), y(_y), z(_z) {}
	NVec3(const NVec2& v);
	NVec3(const NVec2& v, float _z);
	NVec3(const NVec4& v);

	// Member variables
	float x, y, z;

	// Overloaded operators
	NVec3 operator +(const NVec3& v) const	{return NVec3(x + v.x, y + v.y, z + v.z);}
	NVec3 operator -(const NVec3& v) const	{return NVec3(x - v.x, y - v.y, z - v.z);}
	NVec3 operator *(float s) const			{return NVec3(x * s, y * s, z * s);}
	NVec3 operator *(const NVec3& v) const	{return NVec3(x * v.x, y * v.y, z * v.z);}
	NVec3 operator /(float s) const			{float is = 1.0f / s; return NVec3(x * is, y * is, z * is);}
	NVec3 operator /(const NVec3& v) const	{return NVec3(x / v.x, y / v.y, z / v.z);}
	void operator +=(const NVec3& v)			{x += v.x; y += v.y; z += v.z;}
	void operator -=(const NVec3& v)			{x -= v.x; y -= v.y; z -= v.z;}
	void operator *=(float s)				{x *= s; y *= s; z *= s;}
	void operator *=(const NVec3& v)			{x *= v.x; y *= v.y; z *= v.z;}
	void operator /=(float s)				{float is = 1.0f / s; x *= is; y *= is; z *= is;}
	void operator /=(const NVec3& v)			{x /= v.x; y /= v.y; z /= v.z;}

	NVec3 operator *(const NMat4& m) const;
	void operator *=(const NMat4& m);

	// Member functions
	float length() const					{return sqrt(lengthSquared());}
	float lengthSquared() const				{return x*x + y*y + z*z;}
	float length2D() const					{return sqrt(lengthSquared2D());}
	float lengthSquared2D() const			{return x*x + y*y;}
	void normalize()						{float il = 1.0f / length(); il = (il < INFINITY) ? il : 0.0f; x *= il; y *= il; z *= il;}
	void fastNormalize()
	{
		float il = FastInvSqrt((x*x)+(y*y)+(z*z));
		x *= il; y *= il; z *= il;
	}
	void negate()							{x = -x; y = -y; z = -z;}
	NVec3 operator -() const {return Negate(*this);}

private:
};

// Additional functions
inline NVec3 operator *(float s, const NVec3& v) {return v * s;}

inline float Dot(const NVec3& v, const NVec3& w) {return ((v.x * w.x) + (v.y * w.y) + (v.z * w.z));}
inline NVec3 Cross(const NVec3& v, const NVec3& w) {return NVec3(v.y * w.z - v.z * w.y,
															 v.z * w.x - v.x * w.z,
															 v.x * w.y - v.y * w.x);}
inline NVec3 Normalize(const NVec3& v) {float il = 1.0f / v.length(); il = (il < INFINITY) ? il : 0.0f; return NVec3(v.x * il, v.y * il, v.z * il);}
inline NVec3 FastNormalize(const NVec3& v)
{
	float il = FastInvSqrt((v.x*v.x)+(v.y*v.y)+(v.z*v.z));
	return NVec3(v.x * il, v.y * il, v.z * il);
}

inline NVec3 Projection(const NVec3& dir, const NVec3& onto) {if (onto.lengthSquared() > 0.0f) {return onto * (Dot(dir, onto) / Dot(onto, onto));} else {return dir;}}
inline NVec3 Negate(const NVec3& v) {return NVec3(-v.x, -v.y, -v.z);}

inline NVec3 Lerp(const NVec3& a, const NVec3& b, float t)
{
	return NVec3(Lerp(a.x, b.x, t),
				Lerp(a.y, b.y, t),
				Lerp(a.z, b.z, t));
}

inline NVec3 Smooth(const NVec3& a, const NVec3& b, float speed, float deltaTime)
{
	return NVec3(Smooth(a.x, b.x, speed, deltaTime),
				Smooth(a.y, b.y, speed, deltaTime),
				Smooth(a.z, b.z, speed, deltaTime));
}

// Actually a random normal.
inline NVec3 RandomVector()
{
	float theta = random_fl() * PI * 2.0f;
	float phi = acos((2.0f * random_fl()) - 1.0f);

	float x = cos(theta) * sin(phi);
	float y = sin(theta) * sin(phi);
	float z = cos(phi);

	return NVec3(x,y,z);

	//float x = RandomNormal(0.0f, 1.0f);
	//float y = RandomNormal(0.0f, 1.0f);
	//float z = RandomNormal(0.0f, 1.0f);

	//return Normalize(NVec3(x,y,z));
}
