#pragma once

#include <math.h>
#include "NMathDefs.h"
#include "NMathFtns.h"

class NVec2;
class NVec3;
class NMat4;

class NVec4;
inline NVec4 Negate(const NVec4& v);

class NVec4
{
public:
	// Constructors
	NVec4() : x(0.0f), y(0.0f), z(0.0f), w(0.0f) {}
	NVec4(float _x) : x(_x), y(_x), z(_x), w(_x) {}
	NVec4(float _x, float _y, float _z, float _w) : x(_x), y(_y), z(_z), w(_w) {}
	NVec4(const NVec2& v);
	NVec4(const NVec2& v, float _z, float _w);
	NVec4(const NVec3& v);
	NVec4(const NVec3& v, float _w);

	// Member variables
	float x, y, z, w;

	// Overloaded operators
	NVec4 operator +(const NVec4& v) const	{return NVec4(x + v.x, y + v.y, z + v.z, w + v.w);}
	NVec4 operator -(const NVec4& v) const	{return NVec4(x - v.x, y - v.y, z - v.z, w - v.w);}
	NVec4 operator *(float s) const			{return NVec4(x * s, y * s, z * s, w * s);}
	NVec4 operator *(const NVec4& v) const	{return NVec4(x * v.x, y * v.y, z * v.z, w * v.w);}
	NVec4 operator /(float s) const			{float is = 1.0f / s; return NVec4(x * is, y * is, z * is, w * is);}
	NVec4 operator /(const NVec4& v) const	{return NVec4(x / v.x, y / v.y, z / v.z, w / v.w);}
	void operator +=(const NVec4& v)			{x += v.x; y += v.y; z += v.z; w += v.w;}
	void operator -=(const NVec4& v)			{x -= v.x; y -= v.y; z -= v.z; w -= v.w;}
	void operator *=(float s)				{x *= s; y *= s; z *= s; w *= s;}
	void operator *=(const NVec4& v)			{x *= v.x; y *= v.y; z *= v.z; w *= v.w;}
	void operator /=(float s)				{float is = 1.0f / s; x *= is; y *= is; z *= is; w *= is;}
	void operator /=(const NVec4& v)			{x /= v.x; y /= v.y; z /= v.z; w /= v.w;}

	NVec4 operator *(const NMat4& m) const;
	void operator *=(const NMat4& m);

	// Member functions
	float length() const					{return sqrt(lengthSquared());}
	float lengthSquared() const				{return x*x + y*y + z*z + w*w;}
	void normalize()						{float il = 1.0f / length(); il = (il < INFINITY) ? il : 0.0f; x *= il; y *= il; z *= il; w *= il;}
	void fastNormalize()
	{
		float il = FastInvSqrt((x*x)+(y*y)+(z*z)+(w*w));
		x *= il; y *= il; z *= il; w *= il;
	}
	void negate()							{x = -x; y = -y; z = -z; w = -w;}
	NVec4 operator -() const {return Negate(*this);}

private:
};

// Additional functions
inline NVec4 operator *(float s, const NVec4& v) {return v * s;}

inline float Dot(const NVec4& v, const NVec4& w) {return ((v.x * w.x) + (v.y * w.y) + (v.z * w.z) + (v.w * w.w));}

inline NVec4 Normalize(const NVec4& v) {float il = 1.0f / v.length(); il = (il < INFINITY) ? il : 0.0f; return NVec4(v.x * il, v.y * il, v.z * il, v.w * il);}
inline NVec4 FastNormalize(const NVec4& v)
{
	float il = FastInvSqrt((v.x*v.x)+(v.y*v.y)+(v.z*v.z)+(v.w*v.w));
	return NVec4(v.x * il, v.y * il, v.z * il, v.w * il);
}
inline NVec4 Projection(const NVec4& dir, const NVec4& onto) {if (onto.lengthSquared() > 0.0f) {return onto * (Dot(dir, onto) / Dot(onto, onto));} else {return dir;}}
inline NVec4 Negate(const NVec4& v) {return NVec4(-v.x, -v.y, -v.z, -v.w);}

inline NVec4 Lerp(const NVec4& a, const NVec4& b, float t)
{
	return NVec4(Lerp(a.x, b.x, t),
				Lerp(a.y, b.y, t),
				Lerp(a.z, b.z, t),
				Lerp(a.w, b.w, t));
}

inline NVec4 Smooth(const NVec4& a, const NVec4& b, float speed, float deltaTime)
{
	return NVec4(Smooth(a.x, b.x, speed, deltaTime),
				Smooth(a.y, b.y, speed, deltaTime),
				Smooth(a.z, b.z, speed, deltaTime),
				Smooth(a.w, b.w, speed, deltaTime));
}

NVec4 atov(const char* string);
